import socket
import threading
import time
import sys
import signal
import re

from .parser import parse
from .records import SNSrecord
from .exceptions import SNSinteractionException
from .objects import SNSinteraction
from .utils import SingletonMeta, SNS_PORT, BUFFER_SIZE

def signal_handler(signal, frame):
    SNSsession().stop_session_thread = True
    sys.exit(0)

signal.signal(signal.SIGINT, signal_handler)

class SNSsession(metaclass=SingletonMeta):

    def __init__(self):
        self.session_thread = None
        self.session_socket = None
        self.stop_session_thread = True

        self.interactions = []
        self.callbacks = {}

    def start_session(self, address, port=SNS_PORT):
        if self.session_thread is None or not self.session_thread.is_alive():
            self.session_thread = threading.Thread(target=self.__session_thread_func, args=(address, port))
            self.session_thread.start()

    def __session_thread_func(self, address, port=SNS_PORT):
        try:
            self.session_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.session_socket.connect((address, port))

            create_session_interaction = SNSinteraction("USER", "session", {
                'From': 'user:' + SNSrecord().this_user.identifier,
                'To': 'server:' + SNSrecord().this_user.identifier.split('@')[-1]
            })

            self.session_socket.sendall(str(create_session_interaction).encode('utf-8'))

            self.stop_session_thread = False
            while not self.stop_session_thread:
                data = b''

                while True:
                    temp_data = self.session_socket.recv(BUFFER_SIZE)
                    self.receiving_messages = True
                    data += temp_data

                    interaction_blobs = re.split(b'(?:\r?\n){2,}', data)

                    self.append_interaction_blobs(interaction_blobs[:-1])

                    data = interaction_blobs[-1]

                    if len(temp_data) < BUFFER_SIZE: 
                        self.receiving_messages = False
                        break

                    time.sleep(0.1)

                try:
                    self.append_interaction_blobs(re.split(b'(?:\r?\n){2,}', data))
                except SNSinteractionException:
                    print('W: Received interaction not complete')

                self.check_interaction_callbacks()

                time.sleep(0.1)

            self.session_socket.close()
            self.session_socket = None
        except socket.error:
            print("W: Socket error")

        self.stop_session_thread = False

    def append_interaction_blobs(self, interaction_blobs):
        for interaction_blob in interaction_blobs:
            if len(interaction_blob) == 0:
                continue

            try:
                self.interactions.append(parse(interaction_blob.decode('utf-8')))
            except:
                if len(self.interactions) > 0:
                    self.interactions[-1].content.append(interaction_blob.decode('utf-8'))
                else:
                    raise SNSinteractionException(interaction_blob)

    def add_interaction_callback(self, callback_function, type="*", action="*", header_filters={}):
        self.callbacks[type, action] = [header_filters, callback_function]

    def check_interaction_callbacks(self):
        for interaction in self.interactions:
            header_filters = {}
            callback_function = None

            if (interaction.type, interaction.action) in self.callbacks:
                header_filters, callback_function = self.callbacks[interaction.type, interaction.action]
            elif (interaction.type, "*") in self.callbacks:
                header_filters, callback_function = self.callbacks[interaction.type, "*"]
            elif ("*", interaction.action) in self.callbacks:
                header_filters, callback_function = self.callbacks["*", interaction.action]
            elif ("*", "*") in self.callbacks:
                header_filters, callback_function = self.callbacks["*", "*"]
            else:
                print("W: No callback for this interaction, skipping..", interaction.type, interaction.action)
                continue

            for filter_key, filter_value in header_filters.items():
                if filter_key not in interaction.headers or filter_value != interaction.headers[filter_key]:
                    print("W: Interaction does not contain filter headers, skipping..", interaction.type, interaction.action)
                    continue

            callback_function(interaction)

            self.interactions.remove(interaction)

    def close_session(self):
        if not self.session_thread is None and self.session_thread.is_alive():
            self.stop_session_thread = True

            while self.stop_session_thread:
                time.sleep(0.1)

            self.session_thread = None

    def __del__(self):
        self.close_session()
        


from unittest import TestCase

from sns_core.exceptions import SNSparserException, SNSinteractionException, SNSrecordIntegrityException, SNSobjectException, SNSserverException

class TestSNSexceptions(TestCase):

    def test_message_SNSparserException(self):
        exception = SNSparserException('Test', 'SNSparserException')

        assert exception.interaction == 'Test'
        assert exception.message == 'SNSparserException'
        assert str(exception) == 'Test: SNSparserException'

    def test_SNSparserException(self):
        exception = SNSparserException('Test')

        assert exception.interaction == 'Test'
        assert exception.message == 'Exception occurred while parsing the following interaction'
        assert str(exception) == 'Test: Exception occurred while parsing the following interaction'

    def test_message_SNSinteractionException(self):
        exception = SNSinteractionException('Test', 'SNSinteractionException')

        assert exception.interaction == 'Test'
        assert exception.message == 'SNSinteractionException'
        assert str(exception) == 'Test: SNSinteractionException'

    def test_SNSinteractionException(self):
        exception = SNSinteractionException('Test')

        assert exception.interaction == 'Test'
        assert exception.message == 'Exception occurred while requesting with the following interaction'
        assert str(exception) == 'Test: Exception occurred while requesting with the following interaction'

    def test_message_SNSrecordIntegrityException(self):
        exception = SNSrecordIntegrityException('Test', 'SNSrecordIntegrityException')

        assert exception.record == 'Test'
        assert exception.message == 'SNSrecordIntegrityException'
        assert str(exception) == 'Test: SNSrecordIntegrityException'

    def test_SNSrecordIntegrityException(self):
        exception = SNSrecordIntegrityException('Test')

        assert exception.record == 'Test'
        assert exception.message == 'Exception occurred while checked the integrity of the following record'
        assert str(exception) == 'Test: Exception occurred while checked the integrity of the following record'

    def test_message_SNSobjectException(self):
        exception = SNSobjectException('Test', 'SNSobjectException')

        assert exception.object == 'Test'
        assert exception.message == 'SNSobjectException'
        assert str(exception) == 'Test: SNSobjectException'

    def test_SNSobjectException(self):
        exception = SNSobjectException('Test')

        assert exception.object == 'Test'
        assert exception.message == 'Exception occurred while performing an operation on an SNS object'
        assert str(exception) == 'Test: Exception occurred while performing an operation on an SNS object'

    def test_message_SNSserverException(self):
        exception = SNSserverException('SNSserverException')

        assert exception.message == 'SNSserverException'
        assert str(exception) == 'SNSserverException'

    def test_SNSserverException(self):
        exception = SNSserverException()

        assert exception.message == 'Exception occurred while an operation was performed on the server'
        assert str(exception) == 'Exception occurred while an operation was performed on the server'


from unittest import TestCase

from sns_core.records import SNSuserRecordInstance, register_sns_records, check_record_integrity

class TestSNSRecords(TestCase):

    def test_SNSuserRecordInstance(self):
        instance = SNSuserRecordInstance('me', 'sns.provider')

        assert instance.identifier == 'me'
        assert instance.provider == 'sns.provider'


from unittest import TestCase

from sns_core.exceptions import SNSobjectException
from sns_core.objects import SNSinteraction, SNScontent

class TestSNSObjects(TestCase):

    def test_interaction(self):
        interaction = SNSinteraction("USER", "retrieve", headers={'From':'me','To':'you'}, content={'Content':'Hi'})

        self.assertEqual(interaction.type, "USER")
        self.assertEqual(interaction.action, "retrieve")
        self.assertEqual(interaction.headers, {'From':'me',"To":'you'})
        self.assertEqual(type(interaction.content), SNScontent)
        self.assertEqual(interaction.get('From'), 'me')
        self.assertEqual(interaction.get('To'), 'you')
        self.assertEqual(interaction.get('Missing'), None)
        self.assertEqual(interaction.get('Missing', 'Restore'), 'Restore')
        self.assertMultiLineEqual(str(interaction), "USER retrieve\r\nFrom: me\r\nTo: you\r\n\r\nContent: Hi")

    def test_interaction_without_headers(self):
        interaction = SNSinteraction("USER", "retrieve", content=None)

        self.assertEqual(interaction.type, "USER")
        self.assertEqual(interaction.action, "retrieve")
        self.assertEqual(interaction.headers, {})
        self.assertEqual(type(interaction.content), SNScontent)
        self.assertEqual(interaction.get('From'), None)
        self.assertEqual(interaction.get('To'), None)
        self.assertEqual(interaction.get('Missing'), None)
        self.assertEqual(interaction.get('Missing', 'Restore'), 'Restore')
        self.assertMultiLineEqual(str(interaction), "USER retrieve")

    def test_interaction_without_content(self):
        interaction = SNSinteraction("USER", "retrieve", headers={'From':'me','To':'you'})

        self.assertEqual(interaction.type, "USER")
        self.assertEqual(interaction.action, "retrieve")
        self.assertEqual(interaction.headers, {'From':'me',"To":'you'})
        self.assertEqual(type(interaction.content), SNScontent)
        self.assertEqual(interaction.get('From'), 'me')
        self.assertEqual(interaction.get('To'), 'you')
        self.assertEqual(interaction.get('Missing'), None)
        self.assertEqual(interaction.get('Missing', 'Restore'), 'Restore')
        self.assertMultiLineEqual(str(interaction), "USER retrieve\r\nFrom: me\r\nTo: you")

    def test_content_none(self):
        content = SNScontent()

        self.assertEqual(content.content, {})
        self.assertEqual(content.length(), 0)
        self.assertEqual(len(content), 0)
        self.assertEqual(content.get('Missing'), None)
        self.assertEqual(content.get('Missing', 'Restore'), 'Restore')
        self.assertEqual(str(content), '')

    def test_content_list(self):
        with self.assertRaises(SNSobjectException):
            SNScontent([{'Content':'Test'}, {'Content':'Test2'}])

    def test_content_list_all(self):
        with self.assertRaises(SNSobjectException):
            SNScontent([{'Content':['Test','Test2']}, {'Content':'Test3'}, 'Test: yes', ['Test: yes2']])

    def test_content_str(self):
        content = SNScontent('Content: Test')

        self.assertEqual(content.content, {'Content': 'Test'})
        self.assertEqual(content.length(), 1)
        self.assertEqual(len(content), 1)
        self.assertEqual(content.get('Content'), 'Test')
        self.assertEqual(content.get('Content', 'Restore'), 'Test')
        self.assertEqual(content.get('Missing'), None)
        self.assertEqual(content.get('Missing', 'Restore'), 'Restore')
        self.assertEqual(str(content), 'Content: Test')

    def test_content_str_invalid(self):
        with self.assertRaises(SNSobjectException):
            SNScontent('Content:Test')

    def test_content_bytes(self):
        content = SNScontent(b'Content: Test')

        self.assertEqual(content.content, {'Content': 'Test'})
        self.assertEqual(content.length(), 1)
        self.assertEqual(len(content), 1)
        self.assertEqual(content.get('Content'), 'Test')
        self.assertEqual(content.get('Content', 'Restore'), 'Test')
        self.assertEqual(content.get('Missing'), None)
        self.assertEqual(content.get('Missing', 'Restore'), 'Restore')
        self.assertEqual(str(content), 'Content: Test')

    def test_content_bytes_invalid(self):
        with self.assertRaises(SNSobjectException):
            SNScontent(b'0xd1')

    def test_content_dict(self):
        content = SNScontent({'Content': 'Test'})

        self.assertEqual(content.content, {'Content': 'Test'})
        self.assertEqual(content.length(), 1)
        self.assertEqual(len(content), 1)
        self.assertEqual(content.get('Content'), 'Test')
        self.assertEqual(content.get('Content', 'Restore'), 'Test')
        self.assertEqual(content.get('Missing'), None)
        self.assertEqual(content.get('Missing', 'Restore'), 'Restore')
        self.assertEqual(str(content), 'Content: Test')

    def test_content_dict_list(self):
        content = SNScontent({'Content': 'Test'})

        self.assertEqual(content.content, {'Content': 'Test'})
        self.assertEqual(content.length(), 1)
        self.assertEqual(len(content), 1)
        self.assertEqual(content.get('Content'), 'Test')
        self.assertEqual(content.get('Content', 'Restore'), 'Test')
        self.assertEqual(content.get('Missing'), None)
        self.assertEqual(content.get('Missing', 'Restore'), 'Restore')
        self.assertEqual(str(content), 'Content: Test')

    def test_content_SNScontent(self):
        content = SNScontent(SNScontent({'Content': 'Test'}))

        self.assertEqual(content.content, {'Content': 'Test'})
        self.assertEqual(content.length(), 1)
        self.assertEqual(len(content), 1)
        self.assertEqual(content.get('Content'), 'Test')
        self.assertEqual(content.get('Content', 'Restore'), 'Test')
        self.assertEqual(content.get('Missing'), None)
        self.assertEqual(content.get('Missing', 'Restore'), 'Restore')
        self.assertEqual(str(content), 'Content: Test')


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sns_core",
    version="0.0.16",
    author="Soter Company",
    author_email="soter.business@gmail.com",
    description="SNS-core provides all the core functions to communicate using the SNS protocol.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pypi.org/project/sns-core/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    test_suite='tests.test_suite',
    install_requires=[
        
    ],
)


from os.path import expanduser, isdir, isfile
from os import mkdir, mknod, path, makedirs
import json
import sqlite3
from datetime import datetime
from dataclasses import dataclass

from .exceptions import SNSrecordException
from .utils import SingletonMeta, SNS_RECORD_PATH
from . import requests

def register_user(identifier, phone_number):
    # write to file
    if not isdir(expanduser(SNS_RECORD_PATH)):
        makedirs(expanduser(SNS_RECORD_PATH))

    if not isfile(expanduser(SNS_RECORD_PATH) + '/user_record.dat'):
        response = requests.user("create", 'server:0.0.0.0', from_='user:' + identifier, PublicKey='test', PhoneNumber=phone_number)[0]

        if response.action == '500':
            return False

        this_user = UserRecord(-1, response.get('From').split(":")[1], {key: value for key, value in response.headers.items() if 'Key' in key}, response.get('FollowerCount'), response.get('FollowingCount'), response.get('PostCount'), response.get('ReactionCount'))
 
        user_file = open(expanduser(SNS_RECORD_PATH) + '/user_record.dat', 'w')
        json.dump(this_user.toMap(), user_file)
        user_file.close()
    else:
        raise SNSrecordException('User is already registered')

    try:
        register_SNSrecords()
        return True
    except:
        return False

class SNSrecord(metaclass=SingletonMeta):
    def __init__(self, this_user=None, database=None):
        if this_user == None or database == None:
            try:
                if not self.this_user:
                    pass
            except:
                raise SNSrecordException('Record is not initialized')
        else: 
            self.this_user = this_user
            self.database = database
            self.circles = ['private', 'direct', 'group']

    def _check_circle(self, circle):
        if not circle.lower() in self.circles:
            raise SNSrecordException('This circle does not exist')

@dataclass
class UserRecord:
    id: int
    identifier: str
    keys: dict
    follower_count: int
    following_count: int
    post_count: int
    reaction_count: int

    def toMap(self) -> dict:
        return {
            'identifier':self.identifier,
            'keys':json.dumps(self.keys),
            'follower_count':self.follower_count,
            'following_count':self.following_count,
            'post_count':self.post_count,
            'reaction_count':self.reaction_count
        }

@dataclass
class GroupRecord:
    id: int
    identifier: str
    keys: dict
    owners: list
    members: list
    visibility: str
    post_count: int
    reaction_count: int

    def toMap(self) -> dict:
        return {
            'identifier':self.identifier,
            'keys':json.dumps(self.keys),
            'owners':json.dumps(self.owners),
            'members':json.dumps(self.members),
            'visibility':self.visibility,
            'post_count':self.post_count,
            'reaction_count':self.reaction_count
        }

@dataclass
class MessageRecord:
    id: int
    circle: str
    identifier: str
    created_at: datetime
    updated_at: datetime
    content: dict

    def toMap(self) -> dict:
        return {
            'identifier':self.identifier,
            'created_at':int(self.created_at.timestamp() * 1000000),
            'updated_at':int(self.updated_at.timestamp() * 1000000),
            'content':json.dumps(self.content)
        }

@dataclass
class PrivateMessageRecord:

    def toMap(self) -> dict:
        return {
            'identifier':self.identifier,
            'created_at':int(self.created_at.timestamp() * 1000000),
            'updated_at':int(self.updated_at.timestamp() * 1000000),
            'content':json.dumps(self.content)
        }

@dataclass
class DirectMessageRecord:
    user_identifier: str
    is_from: bool

    def toMap(self) -> dict:
        return {
            'identifier':self.identifier,
            'user_identifier':self.user_identifier,
            'from':1 if is_from else 0,
            'created_at':int(self.created_at.timestamp() * 1000000),
            'updated_at':int(self.updated_at.timestamp() * 1000000),
            'content':json.dumps(self.content)
        }

@dataclass
class GroupMessageRecord:
    user_identifier: str
    group_identifier: str

    def toMap(self) -> dict:
        return {
            'identifier':self.identifier,
            'user_identifier':self.user_identifier,
            'group_identifier':self.group_identifier,
            'created_at':int(self.created_at.timestamp() * 1000000),
            'updated_at':int(self.updated_at.timestamp() * 1000000),
            'content':json.dumps(self.content)
        }

def register_SNSrecords():
    global _SNSrecordInstance

    if (not path.isfile(expanduser(SNS_RECORD_PATH) + '/user_record.dat')):
        raise SNSrecordException('The user record file does not exist, indicating this user is not registered')

    file_result = json.load(open(expanduser(SNS_RECORD_PATH) + '/user_record.dat'))

    try:
        _this_user = UserRecord(-1, file_result['identifier'], json.loads(file_result['keys']), 
            file_result['follower_count'], file_result['following_count'],
            file_result['post_count'], file_result['reaction_count'])
    except Exception as e:
        print(e)
        raise SNSrecordException('Missing record item')

    _database = sqlite3.connect(expanduser(SNS_RECORD_PATH) + '/sns_records.db')
    _database.execute('CREATE TABLE IF NOT EXISTS users (id INTEGER PRIMARY KEY AUTOINCREMENT, identifier TEXT UNIQUE, keys TEXT, follower_count INT, following_count INT, post_count INT, reaction_count INT)')
    _database.execute('CREATE TABLE IF NOT EXISTS groups (id INTEGER PRIMARY KEY AUTOINCREMENT, identifier TEXT UNIQUE, keys TEXT, owners TEXT, members TEXT, visibility TEXT, post_count INT, reaction_count INT)')
    _database.execute('CREATE TABLE IF NOT EXISTS private (id INTEGER PRIMARY KEY AUTOINCREMENT, identifier TEXT UNIQUE, created_at INT, updated_at INT, content TEXT)')
    _database.execute('CREATE TABLE IF NOT EXISTS direct (id INTEGER PRIMARY KEY AUTOINCREMENT, identifier TEXT UNIQUE, user_identifier TEXT, [from] INT, created_at INT, updated_at INT, content TEXT)')
    _database.execute('CREATE TABLE IF NOT EXISTS [group] (id INTEGER PRIMARY KEY AUTOINCREMENT, identifier TEXT UNIQUE, user_identifier TEXT, group_identifier TEXT, created_at INT, updated_at INT, content TEXT)')

    SNSrecord(_this_user, _database)